/*
 * Berkeley Lab Checkpoint/Restart (BLCR) for Linux is Copyright (c)
 * 2003, The Regents of the University of California, through Lawrence
 * Berkeley National Laboratory (subject to receipt of any required
 * approvals from the U.S. Dept. of Energy).  All rights reserved.
 *
 * Portions may be copyrighted by others, as may be noted in specific
 * copyright notices within specific files.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: cr_barrier.h,v 1.18 2008/04/11 23:31:47 phargrov Exp $
 */

#ifndef _CR_BARRIER_H
#define _CR_BARRIER_H        1

#include <linux/module.h>
#include <linux/spinlock.h>
#include <linux/wait.h>

// Single-use barrier which can be "notified" by a "third party" (such
// as when a task dies).
//
// A call to cr_barrier_{wait,enter}() will not return until the barrier
// has been passed "count" times.  The term "passing the barrier"
// is taken to mean either a call to cr_barrier_enter() or to
// cr_barrier_notify().
//
// Exactly one call to cr_barrier_{wait,enter,test}{,interruptible}() will return > 0.
// cr_barrier_notify() returns > 0 if it is the "last" arrival.

typedef struct cr_barrier_s {
	wait_queue_head_t	wait;
	atomic_t		count;	// How many remain
	unsigned long		once;
	int			interrupted;
} cr_barrier_t;

extern void cr_barrier_init(cr_barrier_t *barrier, int count);
extern int __cr_barrier_enter(cr_barrier_t *barrier);

/* split-phase treatment */
extern int __cr_barrier_notify(cr_barrier_t *barrier);
extern int cr_barrier_wait(cr_barrier_t *barrier);
extern int cr_barrier_test(cr_barrier_t *barrier);

/* interruptible variants (returns if *any* waiter is interrupted).
 * Upon interruption returns -EINTR to the interrupted task only.
 * The other tasks will wake w/ the normal 0 or 1 return.
 * Not to be mixed w/ non-interruptible variants on the same barrier.
 */
extern int __cr_barrier_enter_interruptible(cr_barrier_t *barrier);
extern int cr_barrier_wait_interruptible(cr_barrier_t *barrier);

/* Debugging wrapper that detects underflow */
#if CRI_DEBUG
    #define _cr_barrier_uflow_wrap(_name, b) ({ \
	cr_barrier_t *_b = (b); \
	int _res = __##_name(_b); \
	if (atomic_read(&_b->count) < 0) \
	    CR_KTRACE_DEBUG(#_name ": underflow on " #b); \
	(_res); \
    })
#else
    #define _cr_barrier_uflow_wrap(_name, b)	__##_name(b)
#endif

/* Tracing wrapper */
#if CR_KERNEL_TRACING
    #define cr_barrier_notify(b) ({ \
	int _res = _cr_barrier_uflow_wrap(cr_barrier_notify, b);\
	CR_KTRACE_BARRIER("NOTIFY(" #b ") returning %d", _res);\
	(_res); \
    })
    #define cr_barrier_enter(b) ({ \
	int _res; \
	CR_KTRACE_BARRIER("ENTER(" #b ")");\
	_res = _cr_barrier_uflow_wrap(cr_barrier_enter, b);\
	CR_KTRACE_BARRIER("LEAVE(" #b ") returning %d", _res);\
	(_res); \
    })
    #define cr_barrier_enter_interruptible(b) ({ \
	int _res; \
	CR_KTRACE_BARRIER("ENTER_INTERRUPTIBLE(" #b ")");\
	_res = _cr_barrier_uflow_wrap(cr_barrier_enter_interruptible, b);\
	CR_KTRACE_BARRIER("LEAVE(" #b ") returning %d", _res);\
	(_res); \
    })
#else
    #define cr_barrier_notify(_b)		_cr_barrier_uflow_wrap(cr_barrier_notify, _b)
    #define cr_barrier_enter(_b)		_cr_barrier_uflow_wrap(cr_barrier_enter, _b)
    #define cr_barrier_enter_interruptible(_b)	_cr_barrier_uflow_wrap(cr_barrier_enter_interruptible, _b)
#endif

/* Macro wrapper to help keep cr_task->step accurate */
#define _cr_barrier_step_wrap(_name, _t, _b) ((_t)->step++, _name(_b))
#define CR_BARRIER_NOTIFY(_t,_b)		_cr_barrier_step_wrap(cr_barrier_notify,_t,_b)
#define CR_BARRIER_ENTER(_t,_b)			_cr_barrier_step_wrap(cr_barrier_enter,_t,_b)
#define CR_BARRIER_ENTER_INTERRUPTIBLE(_t,_b)	_cr_barrier_step_wrap(cr_barrier_enter_interruptible,_t,_b)

#endif
